import tap_tester.connections as connections
import tap_tester.runner as runner
import tap_tester.menagerie as menagerie
from base import ActiveCampaignTest

class ActiveCampaignAllFields(ActiveCampaignTest):
    """Ensure running the tap with all streams and fields selected results in the replication of all fields."""
     
    def name(self):
        return "activecampaign_all_fields"

    def test_run(self):
        """
        • Verify no unexpected streams were replicated
        • Verify that more than just the automatic fields are replicated for each stream. 
        • verify all fields for each stream are replicated
        """
        
        
        # Streams to verify all fields tests
        expected_streams = self.expected_check_streams()

        # We are not able to generate data for `contact_conversions` stream. 
        # For `sms` stream it requires Professional plan of account. So, removing it from streams_to_test set.
        expected_streams = expected_streams - {'contact_conversions', 'sms'}
        
        expected_automatic_fields = self.expected_automatic_fields()
        conn_id = connections.ensure_connection(self)

        found_catalogs = self.run_and_verify_check_mode(conn_id)

        # table and field selection
        test_catalogs_all_fields = [catalog for catalog in found_catalogs
                                    if catalog.get('tap_stream_id') in expected_streams]

        self.perform_and_verify_table_and_field_selection(
            conn_id, test_catalogs_all_fields)

        # grab metadata after performing table-and-field selection to set expectations
        # used for asserting all fields are replicated
        stream_to_all_catalog_fields = dict()
        for catalog in test_catalogs_all_fields:
            stream_id, stream_name = catalog['stream_id'], catalog['stream_name']
            catalog_entry = menagerie.get_annotated_schema(conn_id, stream_id)
            fields_from_field_level_md = [md_entry['breadcrumb'][1]
                                          for md_entry in catalog_entry['metadata']
                                          if md_entry['breadcrumb'] != []]
            stream_to_all_catalog_fields[stream_name] = set(
                fields_from_field_level_md)

        self.run_and_verify_sync(conn_id)

        synced_records = runner.get_records_from_target_output()

        # Verify no unexpected streams were replicated
        synced_stream_names = set(synced_records.keys())
        self.assertSetEqual(expected_streams, synced_stream_names)
        
        
        for stream in expected_streams:
            with self.subTest(stream=stream):

                # expected values
                expected_all_keys = stream_to_all_catalog_fields[stream]
                expected_automatic_keys = expected_automatic_fields.get(
                    stream, set())

                # Verify that more than just the automatic fields are replicated for each stream.
                self.assertTrue(expected_automatic_keys.issubset(
                    expected_all_keys), msg='{} is not in "expected_all_keys"'.format(expected_automatic_keys-expected_all_keys))

                messages = synced_records.get(stream)
                # collect actual values
                actual_all_keys = set()
                for message in messages['messages']:
                    if message['action'] == 'upsert':
                        actual_all_keys.update(message['data'].keys())

                # As we can't generate following field by activecampaign APIs and UI, so removed it form expectation list.
                if stream == "ecommerce_orders":
                    expected_all_keys = expected_all_keys - {'order_products'}
                    
                # verify all fields for each stream are replicated
                self.assertSetEqual(expected_all_keys, actual_all_keys)


import sys
import json
import argparse
import singer
from singer import metadata, utils
from tap_activecampaign.client import ActiveCampaignClient
from tap_activecampaign.discover import discover
from tap_activecampaign.sync import sync

LOGGER = singer.get_logger()

REQUIRED_CONFIG_KEYS = [
    'api_url',
    'api_token',
    'start_date',
    'user_agent'
]

def do_discover():

    LOGGER.info('Starting discover')
    catalog = discover()
    json.dump(catalog.to_dict(), sys.stdout, indent=2)
    LOGGER.info('Finished discover')


@singer.utils.handle_top_exception(LOGGER)
def main():

    parsed_args = singer.utils.parse_args(REQUIRED_CONFIG_KEYS)

    with ActiveCampaignClient(parsed_args.config['api_url'],
                              parsed_args.config['api_token'],
                              parsed_args.config['user_agent'],
                              parsed_args.config.get('request_timeout')) as client:

        state = {}
        if parsed_args.state:
            state = parsed_args.state

        if parsed_args.discover:
            do_discover()
        elif parsed_args.catalog:
            sync(client=client,
                 config=parsed_args.config,
                 catalog=parsed_args.catalog,
                 state=state)

if __name__ == '__main__':
    main()




import singer

from tap_activecampaign.streams import STREAMS, SUB_STREAMS

LOGGER = singer.get_logger()

# Currently syncing sets the stream currently being delivered in the state.
# If the integration is interrupted, this state property is used to identify
#  the starting point to continue from.
# Reference: https://github.com/singer-io/singer-python/blob/master/singer/bookmarks.py#L41-L46
def update_currently_syncing(state, stream_name):
    if (stream_name is None) and ('currently_syncing' in state):
        del state['currently_syncing']
    else:
        singer.set_currently_syncing(state, stream_name)
    singer.write_state(state)


def sync(client, config, catalog, state):
    start_date = config.get('start_date')

    # Get selected_streams from catalog, based on state last_stream
    #   last_stream = Previous currently synced stream, if the load was interrupted
    last_stream = singer.get_currently_syncing(state)
    LOGGER.info('last/currently syncing stream: {}'.format(last_stream))
    selected_streams = []
    for stream in catalog.get_selected_streams(state):
        selected_streams.append(stream.stream)
    LOGGER.info('selected_streams: {}'.format(selected_streams))

    if not selected_streams or selected_streams == []:
        return

    # Loop through endpoints in selected_streams
    for stream_name in selected_streams:

        # parent stream will sync sub stream
        if stream_name in SUB_STREAMS.values():
            continue
        LOGGER.info('START Syncing: {}'.format(stream_name))
        
        stream_obj = STREAMS[stream_name](client)
        stream_obj.write_schema(catalog, stream_name)
        update_currently_syncing(state, stream_name)
        
        total_records = stream_obj.sync(
            client=client,
            catalog=catalog,
            state=state,
            start_date=start_date,
            path=stream_obj.path,
            selected_streams=selected_streams)

        update_currently_syncing(state, None)
        LOGGER.info('FINISHED Syncing: {}, total_records: {}'.format(
            stream_name,
            total_records))



import tap_tester.connections as connections
import tap_tester.runner as runner
from base import ActiveCampaignTest
from datetime import datetime

class ActiveCampaignStartDate(ActiveCampaignTest):
    """
    Ensure both all expected streams respect the start date. Run tap in check mode, 
    run 1st sync with start date = 2021-11-10, run check mode and 2nd sync on a new connection with start date = 13 days later.
    """

    
    start_date_1 = ""
    start_date_2 = ""

    def name(self):
        return "activecampaign_start_date_test"
        
    def test_run(self):
        """
        Test that the start_date configuration is respected
        • verify that a sync with a later start date has at least one record synced
        and less records than the 1st sync with a previous start date
        • verify that each stream has less records than the earlier start date sync
        • verify all data from later start data has bookmark values >= start_date
        """
        # Streams to verify start date tests
        expected_streams = self.expected_check_streams()

        # We are not able to generate data for `contact_conversions`stream. 
        # For `sms` stream it requires Enterprise plan of account. So, removing it from streams_to_test set.
        expected_streams = expected_streams - {'contact_conversions', 'sms'}

        stream_to_test_1 = {'scores', 'contact_data', 'forms', 'templates'}
        self.run_test(days=4, expected_streams= stream_to_test_1)

        stream_to_test_2 = {'conversions', 'conversion_triggers', 'goals', 'site_messages', 'contacts'}
        self.run_test(days=7, expected_streams= stream_to_test_2)
        
        expected_streams = expected_streams - stream_to_test_2 - stream_to_test_1
        self.run_test(days=2, expected_streams=expected_streams)
        
    def run_test(self, days, expected_streams):
        self.start_date_1 = self.get_properties().get('start_date')
        self.start_date_2 = self.timedelta_formatted(self.start_date_1, days=days)
        self.start_date = self.start_date_1

        ##########################################################################
        # First Sync
        ##########################################################################

        # instantiate connection
        conn_id_1 = connections.ensure_connection(self)

        # run check mode
        found_catalogs_1 = self.run_and_verify_check_mode(conn_id_1)

        # table and field selection
        test_catalogs_1_all_fields = [catalog for catalog in found_catalogs_1
                                      if catalog.get('tap_stream_id') in expected_streams]
        self.perform_and_verify_table_and_field_selection(
            conn_id_1, test_catalogs_1_all_fields, select_all_fields=True)

        # run initial sync
        record_count_by_stream_1 = self.run_and_verify_sync(conn_id_1)
        synced_records_1 = runner.get_records_from_target_output()

        ##########################################################################
        # Update START DATE Between Syncs
        ##########################################################################
        
        print("REPLICATION START DATE CHANGE: {} ===>>> {} ".format(
            self.start_date, self.start_date_2))
        self.start_date = self.start_date_2

        ##########################################################################
        # Second Sync
        ##########################################################################

        # create a new connection with the new start_date
        conn_id_2 = connections.ensure_connection(
            self, original_properties=False)

        # run check mode
        found_catalogs_2 = self.run_and_verify_check_mode(conn_id_2)

        # table and field selection
        test_catalogs_2_all_fields = [catalog for catalog in found_catalogs_2
                                      if catalog.get('tap_stream_id') in expected_streams]
        self.perform_and_verify_table_and_field_selection(
            conn_id_2, test_catalogs_2_all_fields, select_all_fields=True)

        # run sync
        record_count_by_stream_2 = self.run_and_verify_sync(conn_id_2)
        synced_records_2 = runner.get_records_from_target_output()

        for stream in expected_streams:
            with self.subTest(stream=stream):

                # expected values
                expected_primary_keys = self.expected_primary_keys()[stream]

                # collect information for assertions from syncs 1 & 2 base on expected values
                record_count_sync_1 = record_count_by_stream_1.get(stream, 0)
                record_count_sync_2 = record_count_by_stream_2.get(stream, 0)

                primary_keys_list_1 = [tuple(message.get('data').get(expected_pk) for expected_pk in expected_primary_keys)
                                       for message in synced_records_1.get(stream, {}).get('messages', [])
                                       if message.get('action') == 'upsert']
                primary_keys_list_2 = [tuple(message.get('data').get(expected_pk) for expected_pk in expected_primary_keys)
                                       for message in synced_records_2.get(stream, {}).get('messages', [])
                                       if message.get('action') == 'upsert']

                primary_keys_sync_1 = set(primary_keys_list_1)
                primary_keys_sync_2 = set(primary_keys_list_2)

                if self.expected_metadata()[stream][self.OBEYS_START_DATE]:

                    # collect information specific to incremental streams from syncs 1 & 2
                    expected_replication_key = next(
                        iter(self.expected_replication_keys().get(stream, [])))
                    replication_dates_1 = [row.get('data').get(expected_replication_key) for row in
                                        synced_records_1.get(stream, {'messages': []}).get('messages', [])
                                        if row.get('data')]
                    replication_dates_2 = [row.get('data').get(expected_replication_key) for row in
                                        synced_records_2.get(stream, {'messages': []}).get('messages', [])
                                        if row.get('data')]

                    # Verify replication key is greater or equal to start_date for sync 1
                    for replication_date in replication_dates_1:

                        self.assertGreaterEqual(
                            self.parse_date(replication_date), self.parse_date(
                                self.start_date_1),
                            msg="Report pertains to a date prior to our start date.\n" +
                            "Sync start_date: {}\n".format(self.start_date_1) +
                                "Record date: {} ".format(replication_date)
                        )

                    # Verify replication key is greater or equal to start_date for sync 2
                    for replication_date in replication_dates_2:

                        self.assertGreaterEqual(
                            self.parse_date(replication_date), self.parse_date(
                                self.start_date_2),
                            msg="Report pertains to a date prior to our start date.\n" +
                            "Sync start_date: {}\n".format(self.start_date_2) +
                                "Record date: {} ".format(replication_date)
                        )
                    # Verify the number of records replicated in sync 1 is greater than the number
                    # of records replicated in sync 2
                    self.assertGreater(record_count_sync_1,
                                       record_count_sync_2)

                    # Verify the records replicated in sync 2 were also replicated in sync 1
                    self.assertTrue(
                        primary_keys_sync_2.issubset(primary_keys_sync_1))
                else:
                    
                    # Verify that the 2nd sync with a later start date replicates the same number of
                    # records as the 1st sync.

                    self.assertEqual(record_count_sync_2, record_count_sync_1)

                    # Verify by primary key the same records are replicated in the 1st and 2nd syncs
                    self.assertSetEqual(primary_keys_sync_1, primary_keys_sync_2)


import re

import tap_tester.connections as connections
from base import ActiveCampaignTest
from tap_tester import menagerie

class ActiveCampaignDiscover(ActiveCampaignTest):
    """
        Testing that discovery creates the appropriate catalog with valid metadata.
        • Verify number of actual streams discovered match expected
        • Verify the stream names discovered were what we expect
        • Verify stream names follow naming convention
          streams should only have lowercase alphas and underscores
        • verify there is only 1 top level breadcrumb
        • verify there are no duplicate metadata entries
        • verify replication key(s)
        • verify primary key(s)
        • verify that if there is a replication key we are doing INCREMENTAL otherwise FULL
        • verify the actual replication matches our expected replication method
        • verify that primary, replication keys are given the inclusion of automatic.
        • verify that all other fields have inclusion of available metadata.
    """

    def name(self):
        return "activecampaign_discover_test"

    def test_run(self):
        streams_to_test = self.expected_check_streams()

        conn_id = connections.ensure_connection(self, payload_hook=None)

        # Verify that there are catalogs found
        found_catalogs = self.run_and_verify_check_mode(
            conn_id)

        # Verify stream names follow naming convention
        # streams should only have lowercase alphas and underscores
        found_catalog_names = {c['tap_stream_id'] for c in found_catalogs}
        self.assertTrue(all([re.fullmatch(r"[a-z_]+",  name) for name in found_catalog_names]),
                        msg="One or more streams don't follow standard naming")
        
        for stream in streams_to_test:
            with self.subTest(stream=stream):

                # Verify the catalog is found for a given stream
                catalog = next(iter([catalog for catalog in found_catalogs
                                     if catalog["stream_name"] == stream]))
                self.assertIsNotNone(catalog)

                # collecting expected values
                expected_primary_keys = self.expected_primary_keys()[stream]
                expected_replication_keys = self.expected_replication_keys()[
                    stream]
                expected_automatic_fields = self.expected_automatic_fields().get(stream)
                expected_replication_method = self.expected_replication_method()[
                    stream]

                # collecting actual values...
                schema_and_metadata = menagerie.get_annotated_schema(
                    conn_id, catalog['stream_id'])
                metadata = schema_and_metadata["metadata"]
                
                stream_properties = [
                    item for item in metadata if item.get("breadcrumb") == []]
                
                actual_primary_keys = set(
                    stream_properties[0].get(
                        "metadata", {self.PRIMARY_KEYS: []}).get(self.PRIMARY_KEYS, [])
                )
                actual_replication_keys = set(
                    stream_properties[0].get(
                        "metadata", {self.REPLICATION_KEYS: []}).get(self.REPLICATION_KEYS, [])
                )
                actual_replication_method = stream_properties[0].get(
                    "metadata", {self.REPLICATION_METHOD: None}).get(self.REPLICATION_METHOD)
                actual_automatic_fields = set(
                    item.get("breadcrumb", ["properties", None])[1] for item in metadata
                    if item.get("metadata").get("inclusion") == "automatic"
                )
                
                actual_fields = []
                for md_entry in metadata:
                    if md_entry['breadcrumb'] != []:
                        actual_fields.append(md_entry['breadcrumb'][1])

                ##########################################################################
                # metadata assertions
                ##########################################################################

                # verify there is only 1 top level breadcrumb in metadata
                self.assertTrue(len(stream_properties) == 1,
                                msg="There is NOT only one top level breadcrumb for {}".format(stream) +
                                "\nstream_properties | {}".format(stream_properties))

                # verify there are no duplicate metadata entries
                self.assertEqual(len(actual_fields), len(set(actual_fields)), msg = f"duplicates in the fields retrieved")

                # verify primary key(s) match expectations
                self.assertSetEqual(
                    expected_primary_keys, actual_primary_keys,
                )
                
                # verify that primary keys and replication keys
                # are given the inclusion of automatic in metadata.
                self.assertSetEqual(expected_automatic_fields,
                                    actual_automatic_fields)

                # verify that all other fields have inclusion of available metadata
                # This assumes there are no unsupported fields for SaaS sources
                self.assertTrue(
                    all({item.get("metadata").get("inclusion") == "available"
                         for item in metadata
                         if item.get("breadcrumb", []) != []
                         and item.get("breadcrumb", ["properties", None])[1]
                         not in actual_automatic_fields}),
                    msg="Not all non key properties are set to available in metadata")

                # verify that if there is a replication key we are doing INCREMENTAL otherwise FULL
                if actual_replication_keys:
                    self.assertTrue(actual_replication_method == self.INCREMENTAL,
                                    msg="Expected INCREMENTAL replication "
                                        "since there is a replication key")
                else:
                    self.assertTrue(actual_replication_method == self.FULL_TABLE,
                                    msg="Expected FULL replication "
                                    "since there is no replication key")

                # verify the actual replication matches our expected replication method
                self.assertEqual(expected_replication_method, actual_replication_method,
                                    msg="The actual replication method {} doesn't match the expected {}".format(
                                        actual_replication_method, expected_replication_method))

                # verify replication key(s) match expectations
                self.assertEqual(expected_replication_keys, actual_replication_keys,
                                 msg="expected replication key {} but actual is {}".format(
                                     expected_replication_keys, actual_replication_keys))


import singer
from singer import metrics, metadata, Transformer, utils
from singer.utils import strptime_to_utc
from tap_activecampaign.transform import transform_json
from tap_activecampaign.client import ActiveCampaignClient

LOGGER = singer.get_logger()
# streams: API URL endpoints to be called
# properties:
#   <root node>: Plural stream name for the endpoint
#   path: API endpoint relative path, when added to the base URL, creates the full path
#   key_properties: Primary key field(s) for the object endpoint
#   replication_method: FULL_TABLE or INCREMENTAL
#   replication_keys: bookmark_field(s), typically a date-time, used for filtering the results
#        and setting the state
#   params: Query, sort, and other endpoint specific parameters
#   data_key: JSON element containing the records for the endpoint
#   bookmark_query_field: Typically a date-time field used for filtering the query
#   bookmark_type: Data type for bookmark, integer or datetime
#   children: A collection of child endpoints (where the endpoint path includes the parent id)
#   parent: On each of the children, the singular stream name for parent element

class ActiveCampaign:
    """
    A base class representing singer streams.
    :param client: The API client used to extract records from external source
    """
    
    stream_name = None
    replication_method = 'INCREMENTAL'
    replication_keys = None
    key_properties = ['id']
    path = None
    params = {}
    parent = None
    data_key = None
    created_timestamp = None
    bookmark_query_field = None
    links = []
    children = []
    
    def __init__(self, client: ActiveCampaignClient = None):
        self.client = client

    def write_schema(self, catalog, stream_name):
        """ 
        Write a schema message.
        """
        stream = catalog.get_stream(stream_name)
        schema = stream.schema.to_dict()
        try:
            # Example:
            # stream = 'test'
            # schema = {'properties': {'id': {'type': 'integer'}, 'email': {'type': 'string'}}}
            # key_properties = ['id']
            # write_schema(stream, schema, key_properties)
            singer.write_schema(stream_name, schema, stream.key_properties)
        except OSError as err:
            LOGGER.error('OS Error while writing schema for: {}'.format(stream_name))
            raise err
        
    def write_record(self, stream_name, record, time_extracted):
        """
        Write a single record for the given stream.
        Example: write_record("users", {"id": 2, "email": "mike@stitchdata.com"})
        """
        try:
            singer.messages.write_record(stream_name, record, time_extracted=time_extracted)
        except OSError as err:
            LOGGER.error('OS Error while writing record for: {}'.format(stream_name))
            LOGGER.error('Stream: {}, record: {}'.format(stream_name, record))
            raise err
        except TypeError as err:
            LOGGER.error('Type Error while writing record for: {}'.format(stream_name))
            LOGGER.error('Stream: {}, record: {}'.format(stream_name, record))
            raise err

    def get_bookmark(self, state, stream, default):
        """ 
        Return bookmark value present in state or return a default value if no bookmark
        present in the state for provided stream
        """
        if (state is None) or ('bookmarks' not in state):
            return default
        return (
            state
            .get('bookmarks', {})
            .get(stream, default)
        )


    def write_bookmark(self, state, stream, value):
        """ Write bookmark in state. """
        if 'bookmarks' not in state:
            state['bookmarks'] = {}
        state['bookmarks'][stream] = value
        LOGGER.info('Write state for stream: {}, value: {}'.format(stream, value))
        singer.write_state(state)

    def transform_datetime(self, this_dttm):
        """
        Transform the datetime to standard datetime format "%Y-%m%dT%H:%M:%S.000000Z"
        """
        with Transformer() as transformer:
            new_dttm = transformer._transform_datetime(this_dttm)
        return new_dttm

    def process_records(self,
                        catalog, #pylint: disable=too-many-branches
                        stream_name,
                        records,
                        time_extracted,
                        bookmark_field=None,
                        max_bookmark_value=None,
                        last_datetime=None,
                        parent=None,
                        parent_id=None):
        """
        This function perform following operation,
        • Transform all the records
        • Update bookmark value to replication key value of record if it is greater than last bookmark value
        • Write only those records of which replication key value is after bookmark value of the last sync for incremental stream.
        • Write all records for FULL_TABLE stream
        • Return updated maximum bookmark value and total count of records
        """
        stream = catalog.get_stream(stream_name)
        schema = stream.schema.to_dict()
        stream_metadata = metadata.to_map(stream.metadata)

        with metrics.record_counter(stream_name) as counter:
            for record in records:
                # If child object, add parent_id to record
                if parent_id and parent:
                    record[parent + '_id'] = parent_id

                # Transform record for Singer.io
                with Transformer() as transformer:
                    try:
                        transformed_record = transformer.transform(
                            record,
                            schema,
                            stream_metadata)
                    except Exception as err:
                        LOGGER.error('Transformer Error: {}'.format(err))
                        LOGGER.error('Stream: {}, record: {}'.format(stream_name, record))
                        raise err

                    # Reset max_bookmark_value to new value if higher
                    if transformed_record.get(bookmark_field):
                        if max_bookmark_value is None or \
                            transformed_record[bookmark_field] > self.transform_datetime(max_bookmark_value):
                            max_bookmark_value = transformed_record[bookmark_field]

                    # If bookmark_field is not none that means stream is incremental.
                    # So, in that case, the tap writes only those records of which the replication key value is greater than last saved bookmark key value
                    # For, FULL_TABLE stream bookmark_field is none. So, in the `else` part it writes all records for the FULL_TABLE stream
                    if bookmark_field and (bookmark_field in transformed_record):
                        last_dttm = self.transform_datetime(last_datetime)
                        bookmark_dttm = self.transform_datetime(transformed_record[bookmark_field])
                        # Keep only records whose bookmark is after the last_datetime
                        if bookmark_dttm:
                            if bookmark_dttm > last_dttm:
                                self.write_record(stream_name, transformed_record, \
                                    time_extracted=time_extracted)
                                counter.increment()
                    else:
                        self.write_record(stream_name, transformed_record, time_extracted=time_extracted)
                        counter.increment()

            # return maximum bookmark value and total no of records
            return max_bookmark_value, counter.value


    # Sync a specific parent or child endpoint.
    def sync(
            self,
            client,
            catalog,
            state,
            start_date,
            path,
            selected_streams=None,
            parent=None,
            parent_id=None):

        static_params = self.params
        bookmark_query_field = self.bookmark_query_field
        bookmark_field = next(iter(self.replication_keys or []), None)
        # Get the latest bookmark for the stream and set the last_integer/datetime
        last_datetime = None
        max_bookmark_value = None

        last_datetime = self.get_bookmark(state, self.stream_name, start_date)
        max_bookmark_value = last_datetime
        LOGGER.info('stream: {}, bookmark_field: {}, last_datetime: {}'.format(
            self.stream_name, bookmark_field, last_datetime))
        now_datetime = utils.now()
        last_dttm = strptime_to_utc(last_datetime)
        endpoint_total = 0

        # pagination: loop thru all pages of data
        # Pagination reference: https://developers.activecampaign.com/reference#pagination
        # Each page has an offset (starting value) and a limit (batch size, number of records)
        # Increase the "offset" by the "limit" for each batch.
        # Continue until the "record_count" returned < "limit" is null/zero or 
        offset = 0 # Starting offset value for each batch API call
        limit = 100 # Batch size; Number of records per API call; Max = 100
        total_records = 0 # Initialize total
        record_count = limit # Initialize, reset for each API call
        page = 1

        while offset <= total_records: # break out of loop when record_count < limit (or not data returned)
            params = {
                'offset': offset,
                'limit': limit,
                **static_params # adds in endpoint specific, sort, filter params
            }

            if bookmark_query_field:
                params[bookmark_query_field] = last_datetime

            # Need URL querystring for 1st page; subsequent pages provided by next_url
            # querystring: Squash query params into string
            querystring = None
            querystring = '&'.join(['%s=%s' % (key, value) for (key, value) in params.items()])

            LOGGER.info('URL for Stream {}: {}{}{}'.format(
                self.stream_name,
                self.client.base_url,
                path,
                '?{}'.format(querystring) if params else ''))

            # API request data
            endpoint_total, total_records, record_count, page, offset, max_bookmark_value = self.get_and_transform_records(
                                querystring, path, max_bookmark_value, state, catalog, start_date, last_datetime, endpoint_total, 
                                  limit, total_records, record_count, page, offset, parent, parent_id, selected_streams)

        # Update the state with the max_bookmark_value for the endpoint
        # ActiveCampaign API does not allow page/batch sorting; bookmark written for endpoint
        if bookmark_field:
            self.write_bookmark(state, self.stream_name, max_bookmark_value)

        # Return total_records (for all pages and date windows)
        return endpoint_total

    def transform_data(self, data):
        """
        Transform data with transform_json from transform.py
        """
        data_key = self.data_key
        
        # The data_key identifies the array/list of records below the <root> element
        transformed_data = [] # initialize the record list
        data_list = []
        data_dict = {}
        if data_key in data:
            if isinstance(data[data_key], list):
                transformed_data = transform_json(data, self.stream_name, data_key)
            elif isinstance(data[data_key], dict):
                data_list.append(data[data_key])
                data_dict[data_key] = data_list
                transformed_data = transform_json(data_dict, self.stream_name, data_key)
        else: # data_key not in data
            if isinstance(data, list):
                data_list = data
                data_dict[data_key] = data_list
                transformed_data = transform_json(data_dict, self.stream_name, data_key)
            elif isinstance(data, dict):
                data_list.append(data)
                data_dict[data_key] = data_list
                transformed_data = transform_json(data_dict, self.stream_name, data_key)
        
        return transformed_data

    def sync_child_stream(self, children, transformed_data, catalog, state, start_date, selected_streams):
        """
        sync the child stream. Loop through all children and if it is selected then collect data based on parent_id.
        """
        id_fields = self.key_properties
        
        for child_stream_name in children:
            if child_stream_name in selected_streams:
                LOGGER.info('START Syncing: {}'.format(child_stream_name))
                child_stream_obj = STREAMS[child_stream_name](self.client)
                child_stream_obj.write_schema(catalog, child_stream_name)
                # For each parent record
                for record in transformed_data:
                    i = 0
                    # Set parent_id
                    for id_field in id_fields:
                        if i == 0:
                            parent_id_field = id_field
                        if id_field == 'id':
                            parent_id_field = id_field
                        i = i + 1
                    parent_id = record.get(parent_id_field)

                    # sync_endpoint for child
                    LOGGER.info(
                        'START Sync for Stream: {}, parent_stream: {}, parent_id: {}'\
                            .format(child_stream_name, self.stream_name, parent_id))
                    child_path = child_stream_obj.path.format(str(parent_id))

                    child_total_records = child_stream_obj.sync(
                        client=self.client,
                        catalog=catalog,
                        state=state,
                        start_date=start_date,
                        path=child_path,
                        selected_streams=selected_streams,
                        parent=child_stream_obj.parent,
                        parent_id=parent_id)
                    LOGGER.info(
                        'FINISHED Sync for Stream: {}, parent_id: {}, total_records: {}'\
                            .format(child_stream_name, parent_id, child_total_records))
                    # End transformed data record loop
                # End if child in selected streams
            # End child streams for parent
        # End if children

    def get_and_transform_records(self, querystring, path, max_bookmark_value, state, catalog, start_date, last_datetime, endpoint_total, 
                                  limit, total_records, record_count, page, offset, parent, parent_id, selected_streams):
        
        """
        Get the records using the client get request and transform it using transform_records
        """
        
        bookmark_field = next(iter(self.replication_keys or []), None)
        created_timestamp_field = self.created_timestamp
        id_fields = self.key_properties

        # API request data
        data = {}
        data = self.client.get(
            path=path,
            params=querystring,
            endpoint=self.stream_name)
        
        # time_extracted: datetime when the data was extracted from the API
        time_extracted = utils.now()
        
        if not data or data is None or data == {}:
            LOGGER.info('No data for URL {}{}{}'.format(self.client.base_url, path, querystring)) # No data results
        else: # has data
            transformed_data = self.transform_data(data)
            
            if not transformed_data or transformed_data is None:
                LOGGER.info('No transformed data for data = {}'.format(data)) # No data results

            i = 0
            for record in transformed_data:
                # Some endpoints update date is null upon creation
                if bookmark_field:
                    created_value = None
                    if created_timestamp_field:
                        created_value = record.get(created_timestamp_field)
                    bookmark_value = record.get(bookmark_field)
                    if not bookmark_value:
                        transformed_data[i][bookmark_field] = created_value
                # Verify key id_fields are present
                for key in id_fields:
                    if not record.get(key):
                        LOGGER.error('Stream: {}, Missing key {} in record: {}'.format(
                            self.stream_name, key, record))
                        raise RuntimeError
                i = i + 1
        
            # Process records and get the max_bookmark_value and record_count for the set of records
            max_bookmark_value, record_count = self.process_records(
                catalog=catalog,
                stream_name=self.stream_name,
                records=transformed_data,
                time_extracted=time_extracted,
                bookmark_field=bookmark_field,
                max_bookmark_value=max_bookmark_value,
                last_datetime=last_datetime,
                parent=parent,
                parent_id=parent_id)
            LOGGER.info('Stream {}, batch processed {} records'.format(
                self.stream_name, record_count))
            endpoint_total = endpoint_total + record_count

            # Loop thru parent batch records for each children objects (if should stream)
            children = self.children

            if children:
                # sync child stream
                self.sync_child_stream(children, transformed_data, catalog, state, start_date, selected_streams)

            # Parent record batch
            # Get pagination details
            api_total = int(data.get('meta', {}).get('total', 0))
            if api_total == 0:
                total_records = record_count
            else:
                total_records = api_total

            # to_rec: to record; ending record for the batch page
            to_rec = offset + limit
            if to_rec > total_records:
                to_rec = total_records

            LOGGER.info('Synced Stream: {}, page: {}, {} to {} of total records: {}'.format(
                self.stream_name,
                page,
                offset,
                to_rec,
                total_records))
            # Pagination: increment the offset by the limit (batch-size) and page
            offset = offset + limit
            page = page + 1
            # End page/batch - while next URL loop

        return endpoint_total, total_records, record_count, page, offset, max_bookmark_value

class Accounts(ActiveCampaign):
    """
    Get data for accounts. 
    Reference : https://developers.activecampaign.com/reference#list-all-accounts
    """
    stream_name = 'accounts'
    replication_keys = ['updated_timestamp']
    path = 'accounts'
    data_key = 'accounts'
    created_timestamp = 'created_timestamp'

class AccountContact(ActiveCampaign):
    """
    Get data for account contact. 
    Reference : https://developers.activecampaign.com/reference#list-all-associations-1
    """
    stream_name = 'account_contacts'
    replication_keys = ['updated_timestamp']
    path = 'accountContacts'
    data_key = 'accountContacts'
    created_timestamp = 'created_timestamp'

class AccountCustomFields(ActiveCampaign):
    """
    Get data for account custom fields. 
    Reference : https://developers.activecampaign.com/reference#list-all-custom-fields
    """
    stream_name = 'account_custom_fields'
    replication_keys = ['updated_timestamp']
    path = 'accountCustomFieldMeta'
    data_key = 'accountCustomFieldMeta'
    created_timestamp = 'created_timestamp'

class AccountCustomFieldValues(ActiveCampaign):
    """
    Get data for account custom field values. 
    Reference : https://developers.activecampaign.com/reference#list-all-custom-field-values-2
    """
    stream_name = 'account_custom_field_values'
    replication_keys = ['updated_timestamp']
    path = 'accountCustomFieldData'
    data_key = 'accountCustomFieldData'
    created_timestamp = 'created_timestamp'

class Addresses(ActiveCampaign):
    """
    Get data for addresses. 
    Reference : https://developers.activecampaign.com/reference#list-all-addresses
    """
    stream_name = 'addresses'
    replication_method = 'FULL_TABLE'
    path = 'addresses'
    data_key = 'addresses'
    links = ['addressGroup', 'addressList']

class Automations(ActiveCampaign):
    """
    Get data for automations. 
    Reference : https://developers.activecampaign.com/reference#list-all-automations
    """
    stream_name = 'automations'
    replication_keys = ['mdate']
    path = 'automations'
    data_key = 'automations'
    created_timestamp = 'cdate'
    links = ['contactGoals', 'blocks']

class Brandings(ActiveCampaign):
    """
    Get data for brandings. 
    Reference : https://developers.activecampaign.com/reference#brandings
    """
    stream_name = 'brandings'
    replication_method = 'FULL_TABLE'
    path = 'brandings'
    data_key = 'brandings'

class Calendars(ActiveCampaign):
    """
    Get data for calendars. 
    Reference : https://developers.activecampaign.com/reference#list-all-calendar-feeds
    """
    stream_name = 'calendars'
    replication_keys = ['mdate']
    path = 'calendars'
    data_key = 'calendars'
    created_timestamp = 'cdate'
    links = ['calendarRels', 'calendarUsers']

class Campaigns(ActiveCampaign):
    """
    Get data for campaigns. 
    Reference : https://developers.activecampaign.com/reference#list-all-campaigns
    """
    stream_name = 'campaigns'
    replication_keys = ['updated_timestamp']
    path = 'campaigns'
    data_key = 'campaigns'
    created_timestamp = 'created_timestamp'

class CampaignLinks(ActiveCampaign):
    """
    Get data for campaign_links. 
    Reference : https://developers.activecampaign.com/reference#retrieve-links-associated-campaign
    """
    stream_name = 'campaign_links'
    replication_keys = ['updated_timestamp']
    path = 'links'
    data_key = 'links'
    created_timestamp = 'created_timestamp'

class Contacts(ActiveCampaign):
    """
    Get data for contacts. 
    Reference : https://developers.activecampaign.com/reference#list-all-contacts
    """
    stream_name = 'contacts'
    replication_keys = ['updated_timestamp']
    path = 'contacts'
    data_key = 'contacts'
    created_timestamp = 'created_timestamp'
    bookmark_query_field = 'filters[updated_after]'
    links = ['contactGoals', 'contactLogs', 'geoIps', 'trackingLogs']

class ContactAutomations(ActiveCampaign):
    """
    Get data for contactAutomations. 
    Reference : https://developers.activecampaign.com/reference#list-all-contact-automations
    """
    stream_name = 'contact_automations'
    replication_keys = ['lastdate']
    path = 'contactAutomations'
    data_key = 'contactAutomations'
    created_timestamp = 'adddate'

class ContactCustomFields(ActiveCampaign):
    """
    Get data for contact_custom_fields. 
    Reference : https://developers.activecampaign.com/reference#retrieve-fields-1
    """
    stream_name = 'contact_custom_fields'
    replication_method = 'FULL_TABLE'
    path = 'fields'
    data_key = 'fields'

class ContactCustomFieldOptions(ActiveCampaign):
    """
    Get data for contact_custom_field_options. 
    Reference : https://developers.activecampaign.com/reference#retrieve-fields-1
    """
    stream_name = 'contact_custom_field_options'
    replication_method = 'FULL_TABLE'
    path = 'fields'
    data_key = 'fieldOptions'

class ContactCustomFieldRels(ActiveCampaign):
    """
    Get data for contact_custom_field_rels. 
    Reference : https://developers.activecampaign.com/reference#retrieve-fields-1
    """
    stream_name = 'contact_custom_field_rels'
    replication_method = 'FULL_TABLE'
    path = 'fields'
    data_key = 'fieldRels'

class ContactCustomFieldValues(ActiveCampaign):
    """
    Get data for contact_custom_field_values. 
    Reference : https://developers.activecampaign.com/reference#list-all-custom-field-values-1
    """
    stream_name = 'contact_custom_field_values'
    replication_keys = ['udate']
    path = 'fieldValues'
    data_key = 'fieldValues'
    created_timestamp = 'cdate'

class ContactDeals(ActiveCampaign):
    """
    Get data for contact_deals. 
    Reference : https://developers.activecampaign.com/reference#list-all-secondary-contacts
    """
    stream_name = 'contact_deals'
    replication_keys = ['updated_timestamp']
    path = 'contactDeals'
    data_key = 'contactDeals'
    created_timestamp = 'created_timestamp'

class DealStages(ActiveCampaign):
    """
    Get data for deal_stages. 
    Reference : https://developers.activecampaign.com/reference#list-all-deal-stages
    """
    stream_name = 'deal_stages'
    replication_keys = ['udate']
    path = 'dealStages'
    data_key = 'dealStages'
    created_timestamp = 'cdate'
    
class DealGroups(ActiveCampaign):
    """
    Get data for deal_groups. 
    Reference : https://developers.activecampaign.com/reference#list-all-pipelines
    Also known as: pipelines
    """
    stream_name = 'deal_groups'
    replication_keys = ['udate']
    path = 'dealGroups'
    data_key = 'dealGroups'
    created_timestamp = 'cdate'
    links = ['dealGroupGroups']

class DealCustomFields(ActiveCampaign):
    """
    Get data for deal_custom_fields. 
    Reference : https://developers.activecampaign.com/reference#retrieve-all-dealcustomfielddata-resources
    """
    stream_name = 'deal_custom_fields'
    replication_keys = ['updated_timestamp']
    path = 'dealCustomFieldMeta'
    data_key = 'dealCustomFieldMeta'
    created_timestamp = 'created_timestamp'

class DealCustomFieldValues(ActiveCampaign):
    """
    Get data for deal_custom_field_values. 
    Reference : https://developers.activecampaign.com/reference#list-all-custom-field-values
    """
    stream_name = 'deal_custom_field_values'
    replication_keys = ['updated_timestamp']
    path = 'dealCustomFieldData'
    data_key = 'dealCustomFieldData'
    created_timestamp = 'created_timestamp'
    
class Deals(ActiveCampaign):
    """
    Get data for deals. 
    Reference : https://developers.activecampaign.com/reference#list-all-deals
    """
    stream_name = 'deals'
    replication_keys = ['mdate']
    path = 'deals'
    data_key = 'deals'
    created_timestamp = 'cdate'

class EcommerceConnections(ActiveCampaign):
    """
    Get data for ecommerce_connections. 
    Reference : https://developers.activecampaign.com/reference#list-all-connections
    """
    stream_name = 'ecommerce_connections'
    
    replication_keys = ['udate']
    path = 'connections'
    data_key = 'connections'
    created_timestamp = 'cdate'
    
class EcommerceCustomers(ActiveCampaign):
    """
    Get data for ecommerce_customers. 
    Reference : https://developers.activecampaign.com/reference#list-all-customers
    """
    stream_name = 'ecommerce_customers'
    replication_keys = ['tstamp']
    path = 'ecomCustomers'
    data_key = 'ecomCustomers'

class EcommerceOrders(ActiveCampaign):
    """
    Get data for ecommerce orders. 
    Reference : https://developers.activecampaign.com/reference#list-all-customers
    """
    stream_name = 'ecommerce_orders'
    replication_keys = ['updated_date']
    path = 'ecomOrders'
    data_key = 'ecomOrders'
    created_timestamp = 'created_timestamp'
    links= ['orderDiscounts']
    children= ['ecommerce_order_products']

class EcommerceOrderProducts(ActiveCampaign):
    """
    Get data for ecommerce order products. 
    Reference : https://developers.activecampaign.com/reference#list-products-for-order
    """
    stream_name = 'ecommerce_order_products'
    replication_method = 'FULL_TABLE'
    path = 'ecomOrders/{}/orderProducts'
    data_key = 'ecomOrderProducts'
    parent = 'ecommerce_orders'

class Forms(ActiveCampaign):
    """
    Get data for forms. 
    Reference : https://developers.activecampaign.com/reference#forms-1
    """
    stream_name = 'forms'
    replication_keys = ['udate']
    path = 'forms'
    data_key = 'forms'
    created_timestamp = 'cdate'

class Groups(ActiveCampaign):
    """
    Get data for groups. 
    Reference : https://developers.activecampaign.com/reference#list-all-groups
    """
    stream_name = 'groups'
    replication_method = 'FULL_TABLE'
    path = 'groups'
    data_key = 'groups'
    
class Lists(ActiveCampaign):
    """
    Get data for lists. 
    Reference : https://developers.activecampaign.com/reference#retrieve-all-lists
    """
    stream_name = 'lists'
    replication_keys = ['updated_timestamp']
    path = 'lists'
    data_key = 'lists'
    created_timestamp = 'created_timestamp'
    links =  ['addressLists', 'contactGoalLists']

class Messages(ActiveCampaign):
    """
    Get data for messages. 
    Reference : https://developers.activecampaign.com/reference#list-all-messages
    """
    stream_name = 'messages'
    replication_keys = ['mdate']
    path = 'messages'
    data_key = 'messages'
    created_timestamp = 'cdate'

class SavedResponses(ActiveCampaign):
    """
    Get data for saved_responses. 
    Reference : https://developers.activecampaign.com/reference#list-all-saved-responses
    """
    stream_name = 'saved_responses'
    replication_keys = ['mdate']
    path = 'savedResponses'
    data_key = 'savedResponses'
    created_timestamp = 'cdate' 

class Scores(ActiveCampaign):
    """
    Get data for scores. 
    Reference : https://developers.activecampaign.com/reference#retrieve-a-score
    """
    stream_name = 'scores'
    replication_keys = ['mdate']
    path = 'scores'
    data_key = 'scores'
    created_timestamp = 'cdate' 

class Segments(ActiveCampaign):
    """
    Get data for segments. 
    Reference : https://developers.activecampaign.com/reference#list-all-segments
    """
    stream_name = 'segments'
    replication_method = 'FULL_TABLE'
    path = 'segments'
    data_key = 'segments'

class Tags(ActiveCampaign):
    """
    Get data for tags. 
    Reference : https://developers.activecampaign.com/reference#list-all-tags
    """
    stream_name = 'tags'
    replication_method = 'FULL_TABLE'
    path = 'tags'
    data_key = 'tags'

class TaskTypes(ActiveCampaign):
    """
    Get data for task_types. 
    Reference : https://developers.activecampaign.com/reference#list-all-deal-task-types
    """
    stream_name = 'task_types'
    replication_method = 'FULL_TABLE'
    path = 'dealTasktypes'
    data_key = 'dealTasktypes'

class Tasks(ActiveCampaign):
    """
    Get data for tasks. 
    Reference : https://developers.activecampaign.com/reference#list-all-tasks
    """
    stream_name = 'tasks'
    replication_keys = ['udate']
    path = 'dealTasks'
    data_key = 'dealTasks'
    created_timestamp = 'cdate'
    links= ['activities', 'taskNotifications']

class Templates(ActiveCampaign):
    """
    Get data for templates. 
    Reference : https://developers.activecampaign.com/reference#templates
    """
    stream_name = 'templates'
    replication_keys = ['mdate']
    path = 'templates'
    data_key = 'templates'

class Users(ActiveCampaign):
    """
    Get data for users. 
    Reference : https://developers.activecampaign.com/reference#users
    """
    stream_name = 'users'
    replication_method = 'FULL_TABLE'
    path = 'users'
    data_key = 'users'

class Webhooks(ActiveCampaign):
    """
    Get data for webhooks. 
    Reference : https://developers.activecampaign.com/reference#webhooks
    """
    stream_name = 'webhooks'
    replication_method = 'FULL_TABLE'
    path = 'webhooks'
    data_key = 'webhooks'

#Undocumented Endpoints

class Activities(ActiveCampaign):
    """
    Get data for activities. 
    """
    stream_name = 'activities'
    replication_keys = ['tstamp']
    path = 'activities'
    data_key = 'activities'

class AutomationBlocks(ActiveCampaign):
    """
    Get data for automation_blocks. 
    """
    stream_name = 'automation_blocks'
    replication_keys = ['mdate']
    path = 'automationBlocks'
    data_key = 'automationBlocks'
    created_timestamp = 'cdate'

class BounceLogs(ActiveCampaign):
    """
    Get data for bounce_logs. 
    """
    stream_name = 'bounce_logs'
    replication_keys = ['updated_timestamp']
    path = 'bounceLogs'
    data_key = 'bounceLogs'
    created_timestamp = 'created_timestamp'

class CampaignLists(ActiveCampaign):
    """
    Get data for campaign_lists. 
    """
    stream_name = 'campaign_lists'
    replication_method = 'FULL_TABLE'
    path = 'campaignLists'
    data_key = 'campaignLists'

class CampaignMessages(ActiveCampaign):
    """
    Get data for campaign_messages. 
    """
    stream_name = 'campaign_messages'
    replication_method = 'FULL_TABLE'
    path = 'campaignMessages'
    data_key = 'campaignMessages'
    
class Configs(ActiveCampaign):
    """
    Get data for configs. 
    """
    stream_name = 'configs'
    replication_keys = ['updated_timestamp']
    path = 'configs'
    data_key = 'configs'
    created_timestamp = 'created_timestamp'

class ContactData(ActiveCampaign):
    """
    Get data for contact_data. 
    """
    stream_name = 'contact_data'
    replication_keys = ['tstamp']
    path = 'contactData'
    data_key = 'contactData'

class ContactEmails(ActiveCampaign):
    """
    Get data for contact_emails. 
    """
    stream_name = 'contact_emails'
    replication_keys = ['sdate']
    path = 'contactEmails'
    data_key = 'contactEmails'


class ContactLists(ActiveCampaign):
    """
    Get data for contact_lists. 
    """
    stream_name = 'contact_lists'
    replication_keys = ['updated_timestamp']
    path = 'contactLists'
    data_key = 'contactLists'
    created_timestamp = 'created_timestamp'

class ContactTags(ActiveCampaign):
    """
    Get data for contact_tags. 
    """
    stream_name = 'contact_tags'
    replication_keys = ['updated_timestamp']
    path = 'contactTags'
    data_key = 'contactTags'
    created_timestamp = 'created_timestamp'

class ContactConversions(ActiveCampaign):
    """
    Get data for contact_conversions. 
    """
    stream_name = 'contact_conversions'
    replication_keys = ['cdate']
    path = 'contactConversions'
    data_key = 'contactConversions'

class Conversions(ActiveCampaign):
    """
    Get data for conversions. 
    """
    stream_name = 'conversions'
    replication_keys = ['udate']
    path = 'conversions'
    data_key = 'conversions'
    created_timestamp = 'cdate'
    links= ['contactConversions']
 
class ConversionTriggers(ActiveCampaign):
    """
    Get data for conversion_triggers. 
    """
    stream_name = 'conversion_triggers'
    replication_keys = ['udate']
    path = 'conversionTriggers'
    data_key = 'conversionTriggers'
    created_timestamp = 'cdate'

class DealActivities(ActiveCampaign):
    """
    Get data for deal_activities. 
    """
    stream_name = 'deal_activities'
    replication_keys = ['cdate']
    path = 'dealActivities'
    data_key = 'dealActivities'

class DealGroupUsers(ActiveCampaign):
    """
    Get data for deal_group_users. 
    """
    stream_name = 'deal_group_users'
    replication_method = 'FULL_TABLE'
    path = 'dealGroupUsers'
    data_key = 'dealGroupUsers'

class EcommerceOrderActivities(ActiveCampaign):
    """
    Get data for ecommerce_order_activities. 
    """
    stream_name = 'ecommerce_order_activities'
    replication_keys = ['updated_date']
    path = 'ecomOrderActivities'
    data_key = 'ecomOrderActivities'
    created_timestamp= 'created_date',

class EmailActivities(ActiveCampaign):
    """
    Get data for email_activities. 
    """
    stream_name = 'email_activities'
    replication_keys = ['tstamp']
    path = 'emailActivities'
    data_key = 'emailActivities'

class Goals(ActiveCampaign):
    """
    Get data for goals. 
    """
    stream_name = 'goals'
    replication_method = 'FULL_TABLE'
    path = 'goals'
    data_key = 'goals'

class SiteMessages(ActiveCampaign):
    """
    Get data for site_messages. 
    """
    stream_name = 'site_messages'
    replication_keys = ['ldate']
    path = 'siteMessages'
    data_key = 'siteMessages'

class Sms(ActiveCampaign):
    """
    Get data for sms. 
    """
    stream_name = 'sms'
    replication_keys = ['tstamp']
    path = 'sms'
    data_key = 'sms'

STREAMS = {
  'accounts': Accounts,
  'account_contacts': AccountContact,
  'account_custom_fields': AccountCustomFields,
  'account_custom_field_values': AccountCustomFieldValues,
  'addresses': Addresses,
  'automations': Automations,
  'brandings': Brandings,
  'calendars': Calendars,
  'campaigns': Campaigns,
  'campaign_links': CampaignLinks,
  'contacts': Contacts,
  'contact_automations': ContactAutomations,
  'contact_custom_fields': ContactCustomFields,
  'contact_custom_field_options': ContactCustomFieldOptions,
  'contact_custom_field_rels': ContactCustomFieldRels,
  'contact_custom_field_values': ContactCustomFieldValues,
  'contact_deals': ContactDeals,
  'deal_stages': DealStages,
  'deal_groups': DealGroups,
  'deal_custom_fields': DealCustomFields,
  'deal_custom_field_values': DealCustomFieldValues,
  'deals': Deals,
  'ecommerce_connections': EcommerceConnections,
  'ecommerce_customers': EcommerceCustomers,
  'ecommerce_orders': EcommerceOrders,
  'ecommerce_order_products': EcommerceOrderProducts,
  'forms': Forms,
  'groups': Groups,
  'lists': Lists,
  'messages': Messages,
  'saved_responses': SavedResponses,
  'scores': Scores,
  'segments': Segments,
  'tags': Tags,
  'task_types': TaskTypes,
  'tasks': Tasks,
  'templates': Templates,
  'users': Users,
  'webhooks': Webhooks,
  'activities': Activities,
  'automation_blocks': AutomationBlocks,
  'bounce_logs': BounceLogs,
  'campaign_lists': CampaignLists,
  'campaign_messages': CampaignMessages,
  'configs': Configs,
  'contact_data': ContactData,
  'contact_emails': ContactEmails,
  'contact_lists': ContactLists,
  'contact_tags': ContactTags,
  'contact_conversions': ContactConversions,
  'conversions': Conversions,
  'conversion_triggers': ConversionTriggers,
  'deal_activities': DealActivities,
  'deal_group_users': DealGroupUsers,
  'ecommerce_order_activities': EcommerceOrderActivities,
  'email_activities': EmailActivities,
  'goals': Goals,
  'site_messages': SiteMessages,
  'sms': Sms
}

SUB_STREAMS = {
    'ecommerce_orders': 'ecommerce_order_products'
}

def flatten_streams():
    flat_streams = {}
    # Loop through all streams
    for stream in STREAMS.values():
      stream = stream()
      flat_streams[stream.stream_name] = {
            'key_properties': stream.key_properties,
            'replication_method': stream.replication_method,
            'replication_keys': stream.replication_keys
        }

    return flat_streams



import backoff
import requests
from singer import metrics, utils
import singer

LOGGER = singer.get_logger()
REQUEST_TIMEOUT = 300

DEFAULT_API_VERSION = '3'


class Server5xxError(Exception):
    pass


class Server429Error(Exception):
    pass


class ActiveCampaignError(Exception):
    pass
class ActiveCampaignBadRequestError(ActiveCampaignError):
    pass

class ActiveCampaignUnauthorizedError(ActiveCampaignError):
    pass

class ActiveCampaignForbiddenError(ActiveCampaignError):
    pass

class ActiveCampaignNotFoundError(ActiveCampaignError):
    pass

class ActiveCampaignUnprocessableEntityError(ActiveCampaignError):
    pass

class ActiveCampaignRateLimitError(Server429Error):
    pass

class ActiveCampaignInternalServerError(Server5xxError):
    pass


# Errors Reference: https://developers.activecampaign.com/reference#errors
STATUS_CODE_EXCEPTION_MAPPING = {
    400: {
        "raise_exception": ActiveCampaignBadRequestError,
        "message": "A validation exception has occurred."
    },
    401: {
        "raise_exception": ActiveCampaignUnauthorizedError,
        "message": "Invalid authorization credentials."
    },
    403: {
        "raise_exception": ActiveCampaignForbiddenError,
        "message": "The request could not be authenticated or the authenticated user is not authorized to access the requested resource."
    },
    404: {
        "raise_exception": ActiveCampaignNotFoundError,
        "message": "The requested resource does not exist."
    },
    422: {
        "raise_exception": ActiveCampaignUnprocessableEntityError,
        "message": "The request could not be processed, usually due to a missing or invalid parameter."
    },
    429: {
        "raise_exception": ActiveCampaignRateLimitError,
        "message": "The user has sent too many requests in a given amount of time ('rate limiting') - contact support or account manager for more details."
    },
    500: {
        "raise_exception": ActiveCampaignInternalServerError,
        "message": "The server encountered an unexpected condition which prevented" \
            " it from fulfilling the request."
    }
}

def should_retry_error(exception):
    """ 
        Return true if exception is required to retry otherwise return false
    """

    if isinstance(exception, OSError) or isinstance(exception, Server5xxError) or isinstance(exception, Server429Error):
        # Retry Server5xxError and Server429Error exception. Retry exception if it is child class of OSError.
        # OSError is Parent class of ConnectionError, ConnectionResetError, TimeoutError and other errors mentioned in https://docs.python.org/3/library/exceptions.html#os-exceptions
        return True
    elif type(exception) == Exception and type(exception.args[0][1]) == ConnectionResetError:
        # Tap raises Exception: ConnectionResetError(104, 'Connection reset by peer'). That's why we are retrying this error also.
        # Reference: https://app.circleci.com/pipelines/github/singer-io/tap-activecampaign/554/workflows/d448258e-20df-4e66-b2aa-bc8bd1f08912/jobs/558
        return True
    else:
        return False

def get_exception_for_status_code(status_code):
    # Map the status code with `STATUS_CODE_EXCEPTION_MAPPING` dictionary and accordingly return the error.
    if status_code > 500:
        # Raise Server5xxError if status code is greater than 500
        return Server5xxError

    return STATUS_CODE_EXCEPTION_MAPPING.get(status_code, {}).get("raise_exception", ActiveCampaignError)

# Example 422 error
# {
#   "errors": [
#     {
#       "title": "The connection service was not provided."
#     },
#     {
#       "title": "The connection externalid was not provided."
#     }
#   ]
# }
def raise_for_error(response):
    # If the response contains an `errors` object with multiple errors, prepare an error message containing the title of all errors in the `errors` object.
    # Otherwise prepare a custom default message and raise the exception.

    try:
        response_json = response.json() # retrieve json response
    except Exception:
        response_json = {}
    errors = response_json.get('errors', [])
    status_code = response.status_code
    if errors: # response containing `errors` object
        message = 'HTTP-error-code: {}, Error:'.format(status_code)
        # loop through all errors
        for error in errors:
            title = error.get('title')
            message = '{} {}'.format(message, title)
    else:
        # prepare custom default error message
        message = "HTTP-error-code: {}, Error: {}".format(status_code,
                response_json.get("message", STATUS_CODE_EXCEPTION_MAPPING.get(
                status_code, {}).get("message", "Unknown Error")))
    
    exc = get_exception_for_status_code(status_code)

    raise exc(message) from None


class ActiveCampaignClient(object):
    def __init__(self,
                 api_url,
                 api_token,
                 user_agent=None,
                 request_timeout=None):
        self.__api_url = api_url
        self.__api_token = api_token
        self.__user_agent = user_agent
        self.__session = requests.Session()
        self.__verified = False
        self.base_url = '{}/api/{}/'.format(self.__api_url, DEFAULT_API_VERSION)

        # if request_timeout is other than 0, "0" or "" then use request_timeout
        if request_timeout and float(request_timeout):
            self.request_timeout = float(request_timeout)
        else: # If value is 0, "0" or "" then set default to 300 seconds.
            self.request_timeout = REQUEST_TIMEOUT

    # Backoff for Server5xxError, Server429Error, OSError and Exception with ConnectionResetError.
    @backoff.on_exception(backoff.expo,
                          (Exception),
                          giveup=lambda e: not should_retry_error(e),
                          max_tries=5,
                          factor=2)
    def __enter__(self):
        self.__verified = self.check_api_token()
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        self.__session.close()

    def check_api_token(self):
        if self.__api_token is None:
            raise Exception('Error: Missing api_token.')
        headers = {}
        if self.__user_agent:
            headers['User-Agent'] = self.__user_agent
        headers['Api-Token'] = self.__api_token
        headers['Accept'] = 'application/json'
        url = self.base_url
        response = self.__session.get(
            # Simple endpoint that returns 1 record w/ default organization URN
            url=url,
            headers=headers,
            timeout=self.request_timeout)
        if response.status_code != 200:
            raise_for_error(response)
        else:
            return True

    # Backoff for Server5xxError, Server429Error, OSError and Exception with ConnectionResetError.
    @backoff.on_exception(backoff.expo,
                          (Exception),
                          giveup=lambda e: not should_retry_error(e),
                          max_tries=5,
                          factor=2)
    # Rate limit: https://developers.activecampaign.com/reference#rate-limits
    @utils.ratelimit(5, 1)
    def request(self, method, path=None, url=None, api_version=None, **kwargs):
        if not self.__verified:
            self.__verified = self.check_api_token()

        if not api_version:
            api_version = DEFAULT_API_VERSION

        if not url and path:
            url = '{}/api/{}/{}'.format(self.__api_url, api_version, path)

        if 'endpoint' in kwargs:
            endpoint = kwargs['endpoint']
            del kwargs['endpoint']
        else:
            endpoint = None

        if 'headers' not in kwargs:
            kwargs['headers'] = {}
        kwargs['headers']['Api-Token'] = self.__api_token
        kwargs['headers']['Accept'] = 'application/json'

        if self.__user_agent:
            kwargs['headers']['User-Agent'] = self.__user_agent

        if method == 'POST':
            kwargs['headers']['Content-Type'] = 'application/json'

        with metrics.http_request_timer(endpoint) as timer:
            response = self.__session.request(method, url, stream=True, timeout=self.request_timeout, **kwargs)
            timer.tags[metrics.Tag.http_status_code] = response.status_code

        if response.status_code != 200:
            raise_for_error(response)

        # Log invalid JSON (e.g. unterminated string errors)
        try:
            response_json = response.json()
        except Exception as err:
            LOGGER.error('{}'.format(err))
            LOGGER.error('response content: {}'.format(response.content))
            raise Exception(err)

        return response_json

    def get(self, path, api_version=None, **kwargs):
        return self.request('GET', path=path, api_version=api_version, **kwargs)

    def post(self, path, api_version=None, **kwargs):
        return self.request('POST', path=path, api_version=api_version, **kwargs)


import singer
from singer.catalog import Catalog, CatalogEntry, Schema
from tap_activecampaign.schema import get_schemas
from tap_activecampaign.streams import flatten_streams

LOGGER = singer.get_logger()

def discover():
    schemas, field_metadata = get_schemas()
    catalog = Catalog([])

    flat_streams = flatten_streams()
    for stream_name, schema_dict in schemas.items():
        try:
            schema = Schema.from_dict(schema_dict)
            mdata = field_metadata[stream_name]
        except Exception as err:
            LOGGER.error(err)
            LOGGER.error('stream_name: {}'.format(stream_name))
            LOGGER.error('type schema_dict: {}'.format(type(schema_dict)))
            raise err

        catalog.streams.append(CatalogEntry(
            stream=stream_name,
            tap_stream_id=stream_name,
            key_properties=flat_streams.get(stream_name, {}).get('key_properties', None),
            schema=schema,
            metadata=mdata
        ))

    return catalog



import unittest
import os
import tap_tester.connections as connections
import tap_tester.menagerie as menagerie
import tap_tester.runner as runner
from datetime import datetime as dt
from datetime import timedelta
import dateutil.parser
import pytz

class ActiveCampaignTest(unittest.TestCase):
    start_date = ""
    
    START_DATE_FORMAT = "%Y-%m-%dT00:00:00Z"
    PRIMARY_KEYS = "table-key-properties"
    REPLICATION_METHOD = "forced-replication-method"
    REPLICATION_KEYS = "valid-replication-keys"
    FULL_TABLE = "FULL_TABLE"
    INCREMENTAL = "INCREMENTAL"
    OBEYS_START_DATE = "obey-start-date"

    def tap_name(self):
        """The name of the tap"""
        return "tap-activecampaign"
    
    def setUp(self):
        required_env = {
            "TAP_ACTIVECAMPAIGN_API_TOKEN",
            "TAP_ACTIVECAMPAIGN_API_URL",
        }
        missing_envs = [v for v in required_env if not os.getenv(v)]
        if missing_envs:
            raise Exception("set " + ", ".join(missing_envs))
        
    def get_type(self):
        return "platform.activecampaign"

    def get_credentials(self):
        """Authentication information for the test account"""
        return {
            'api_url': os.getenv('TAP_ACTIVECAMPAIGN_API_URL'),
            'api_token': os.getenv('TAP_ACTIVECAMPAIGN_API_TOKEN')
        }

    def get_properties(self, original: bool = True):
        """Configuration properties required for the tap."""
        return_value = {
            "start_date" : "2021-12-01T00:00:00Z",
        }
        if original:
            return return_value

        # Reassign start date
        return_value["start_date"] = self.start_date
        return return_value
    
    def expected_metadata(self):
        """The expected streams and metadata about the streams"""
        return {
            'accounts': {
                self.PRIMARY_KEYS: {'id'},
                self.REPLICATION_METHOD: self.INCREMENTAL,
                self.REPLICATION_KEYS: {'updated_timestamp'},
                self.OBEYS_START_DATE: True
            },
            'account_contacts': {
                self.PRIMARY_KEYS: {'id'},
                self.REPLICATION_METHOD: self.INCREMENTAL,
                self.REPLICATION_KEYS: {'updated_timestamp'},
                self.OBEYS_START_DATE: True      
            },
            'account_custom_fields': {
                self.PRIMARY_KEYS: {'id'},
                self.REPLICATION_METHOD: self.INCREMENTAL,
                self.REPLICATION_KEYS: {'updated_timestamp'},
                self.OBEYS_START_DATE: True
            },
            'account_custom_field_values': {
                self.PRIMARY_KEYS: {'id'},
                self.REPLICATION_METHOD: self.INCREMENTAL,
                self.REPLICATION_KEYS: {'updated_timestamp'},
                self.OBEYS_START_DATE: True
            },
            'addresses': {
                self.PRIMARY_KEYS: {'id'},
                self.REPLICATION_METHOD: self.FULL_TABLE,
                self.OBEYS_START_DATE: False
            },
            'automations': {
                self.PRIMARY_KEYS: {'id'},
                self.REPLICATION_METHOD: self.INCREMENTAL,
                self.REPLICATION_KEYS: {'mdate'},
                self.OBEYS_START_DATE: True
            },
            'brandings': {
                self.PRIMARY_KEYS: {'id'},
                self.REPLICATION_METHOD: self.FULL_TABLE,
                self.OBEYS_START_DATE: False
            },
            'calendars': {
                self.PRIMARY_KEYS: {'id'},
                self.REPLICATION_METHOD: self.INCREMENTAL,
                self.REPLICATION_KEYS: {'mdate'},
                self.OBEYS_START_DATE: True
            },
            'campaigns': {
                self.PRIMARY_KEYS: {'id'},
                self.REPLICATION_METHOD: self.INCREMENTAL,
                self.REPLICATION_KEYS: {'updated_timestamp'},
                self.OBEYS_START_DATE: True
            },
            'campaign_links': {
                self.PRIMARY_KEYS: {'id'},
                self.REPLICATION_METHOD: self.INCREMENTAL,
                self.REPLICATION_KEYS: {'updated_timestamp'},
                self.OBEYS_START_DATE: True
            },
            'contacts': {
                self.PRIMARY_KEYS: {'id'},
                self.REPLICATION_METHOD: self.INCREMENTAL,
                self.REPLICATION_KEYS: {'updated_timestamp'},
                self.OBEYS_START_DATE: True
            },
            'contact_automations': {
                self.PRIMARY_KEYS: {'id'},
                self.REPLICATION_METHOD: self.INCREMENTAL,
                self.REPLICATION_KEYS: {'lastdate'},
                self.OBEYS_START_DATE: True
            },
            'contact_custom_fields': {
                self.PRIMARY_KEYS: {'id'},
                self.REPLICATION_METHOD: self.FULL_TABLE,
                self.OBEYS_START_DATE: False
            },
            'contact_custom_field_options': {
                self.PRIMARY_KEYS: {'id'},
                self.REPLICATION_METHOD: self.FULL_TABLE,
                self.OBEYS_START_DATE: False
            },
            'contact_custom_field_rels': {
                self.PRIMARY_KEYS: {'id'},
                self.REPLICATION_METHOD: self.FULL_TABLE,
                self.OBEYS_START_DATE: False
            },
            'contact_custom_field_values': {
                self.PRIMARY_KEYS: {'id'},
                self.REPLICATION_METHOD: self.INCREMENTAL,
                self.REPLICATION_KEYS: {'udate'},
                self.OBEYS_START_DATE: True
            },
            'contact_deals': {
                self.PRIMARY_KEYS: {'id'},
                self.REPLICATION_METHOD: self.INCREMENTAL,
                self.REPLICATION_KEYS: {'updated_timestamp'},
                self.OBEYS_START_DATE: True
            },
            'deal_stages': {
                self.PRIMARY_KEYS: {'id'},
                self.REPLICATION_METHOD: self.INCREMENTAL,
                self.REPLICATION_KEYS: {'udate'},
                self.OBEYS_START_DATE: True
            },
            'deal_groups': {
                self.PRIMARY_KEYS: {'id'},
                self.REPLICATION_METHOD: self.INCREMENTAL,
                self.REPLICATION_KEYS: {'udate'},
                self.OBEYS_START_DATE: True
            },
            'deal_custom_fields': {
                self.PRIMARY_KEYS: {'id'},
                self.REPLICATION_METHOD: self.INCREMENTAL,
                self.REPLICATION_KEYS: {'updated_timestamp'},
                self.OBEYS_START_DATE: True
            },
            'deal_custom_field_values': {
                self.PRIMARY_KEYS: {'id'},
                self.REPLICATION_METHOD: self.INCREMENTAL,
                self.REPLICATION_KEYS: {'updated_timestamp'},
                self.OBEYS_START_DATE: True
            },
            'deals': {
                self.PRIMARY_KEYS: {'id'},
                self.REPLICATION_METHOD: self.INCREMENTAL,
                self.REPLICATION_KEYS: {'mdate'},
                self.OBEYS_START_DATE: True
            },
            'ecommerce_connections': {
                self.PRIMARY_KEYS: {'id'},
                self.REPLICATION_METHOD: self.INCREMENTAL,
                self.REPLICATION_KEYS: {'udate'},
                self.OBEYS_START_DATE: True
            },
            'ecommerce_customers': {
                self.PRIMARY_KEYS: {'id'},
                self.REPLICATION_METHOD: self.INCREMENTAL,
                self.REPLICATION_KEYS: {'tstamp'},
                self.OBEYS_START_DATE: True
            },
            'ecommerce_orders': {
                self.PRIMARY_KEYS: {'id'},
                self.REPLICATION_METHOD: self.INCREMENTAL,
                self.REPLICATION_KEYS: {'updated_date'},
                self.OBEYS_START_DATE: True
            },
            'ecommerce_order_products': {
                self.PRIMARY_KEYS: {'id'},
                self.REPLICATION_METHOD: self.FULL_TABLE,
                self.OBEYS_START_DATE: False
            },
            'forms': {
                self.PRIMARY_KEYS: {'id'},
                self.REPLICATION_METHOD: self.INCREMENTAL,
                self.REPLICATION_KEYS: {'udate'},
                self.OBEYS_START_DATE: True
            },
            'groups': {
                self.PRIMARY_KEYS: {'id'},
                self.REPLICATION_METHOD: self.FULL_TABLE,
                self.OBEYS_START_DATE: False
            },
            'lists': {
                self.PRIMARY_KEYS: {'id'},
                self.REPLICATION_METHOD: self.INCREMENTAL,
                self.REPLICATION_KEYS: {'updated_timestamp'},
                self.OBEYS_START_DATE: True
            },
            'messages': {
                self.PRIMARY_KEYS: {'id'},
                self.REPLICATION_METHOD: self.INCREMENTAL,
                self.REPLICATION_KEYS: {'mdate'},
                self.OBEYS_START_DATE: True
            },
            'saved_responses': {
                self.PRIMARY_KEYS: {'id'},
                self.REPLICATION_METHOD: self.INCREMENTAL,
                self.REPLICATION_KEYS: {'mdate'},
                self.OBEYS_START_DATE: True
            },
            'scores': {
                self.PRIMARY_KEYS: {'id'},
                self.REPLICATION_METHOD: self.INCREMENTAL,
                self.REPLICATION_KEYS: {'mdate'},
                self.OBEYS_START_DATE: True
            },
            'segments': {
                self.PRIMARY_KEYS: {'id'},
                self.REPLICATION_METHOD: self.FULL_TABLE,
                self.OBEYS_START_DATE: False
            },
            'tags': {
                self.PRIMARY_KEYS: {'id'},
                self.REPLICATION_METHOD: self.FULL_TABLE,
                self.OBEYS_START_DATE: False
            },
            'task_types': {
                self.PRIMARY_KEYS: {'id'},
                self.REPLICATION_METHOD: self.FULL_TABLE,
                self.OBEYS_START_DATE: False
            },
            'tasks': {
                self.PRIMARY_KEYS: {'id'},
                self.REPLICATION_METHOD: self.INCREMENTAL,
                self.REPLICATION_KEYS: {'udate'},
                self.OBEYS_START_DATE: True
            },
            'templates': {
                self.PRIMARY_KEYS: {'id'},
                self.REPLICATION_METHOD: self.INCREMENTAL,
                self.REPLICATION_KEYS: {'mdate'},
                self.OBEYS_START_DATE: True
            },
            'users': {
                self.PRIMARY_KEYS: {'id'},
                self.REPLICATION_METHOD: self.FULL_TABLE,
                self.OBEYS_START_DATE: False
            },
            'webhooks': {
                self.PRIMARY_KEYS: {'id'},
                self.REPLICATION_METHOD: self.FULL_TABLE,
                self.OBEYS_START_DATE: False
            },
            'activities': {
                self.PRIMARY_KEYS: {'id'},
                self.REPLICATION_METHOD: self.INCREMENTAL,
                self.REPLICATION_KEYS: {'tstamp'},
                self.OBEYS_START_DATE: True
            },
            'automation_blocks': {
                self.PRIMARY_KEYS: {'id'},
                self.REPLICATION_METHOD: self.INCREMENTAL,
                self.REPLICATION_KEYS: {'mdate'},
                self.OBEYS_START_DATE: True
            },
            'bounce_logs': {
                self.PRIMARY_KEYS: {'id'},
                self.REPLICATION_METHOD: self.INCREMENTAL,
                self.REPLICATION_KEYS: {'updated_timestamp'},
                self.OBEYS_START_DATE: True
            },
            'campaign_lists': {
                self.PRIMARY_KEYS: {'id'},
                self.REPLICATION_METHOD: self.FULL_TABLE,
                self.OBEYS_START_DATE: False
            },
            'campaign_messages': {
                self.PRIMARY_KEYS: {'id'},
                self.REPLICATION_METHOD: self.FULL_TABLE,
                self.OBEYS_START_DATE: False
            },
            'configs': {
                self.PRIMARY_KEYS: {'id'},
                self.REPLICATION_METHOD: self.INCREMENTAL,
                self.REPLICATION_KEYS: {'updated_timestamp'},
                self.OBEYS_START_DATE: True
            },
            'contact_data': {
                self.PRIMARY_KEYS: {'id'},
                self.REPLICATION_METHOD: self.INCREMENTAL,
                self.REPLICATION_KEYS: {'tstamp'},
                self.OBEYS_START_DATE: True
            },
            'contact_emails': {
                self.PRIMARY_KEYS: {'id'},
                self.REPLICATION_METHOD: self.INCREMENTAL,
                self.REPLICATION_KEYS: {'sdate'},
                self.OBEYS_START_DATE: True
            },
            'contact_lists': {
                self.PRIMARY_KEYS: {'id'},
                self.REPLICATION_METHOD: self.INCREMENTAL,
                self.REPLICATION_KEYS: {'updated_timestamp'},
                self.OBEYS_START_DATE: True
            },
            'contact_tags': {
                self.PRIMARY_KEYS: {'id'},
                self.REPLICATION_METHOD: self.INCREMENTAL,
                self.REPLICATION_KEYS: {'updated_timestamp'},
                self.OBEYS_START_DATE: True
            },
            'contact_conversions': {
                self.PRIMARY_KEYS: {'id'},
                self.REPLICATION_METHOD: self.INCREMENTAL,
                self.REPLICATION_KEYS: {'cdate'},
                self.OBEYS_START_DATE: True
            },
            'conversions': {
                self.PRIMARY_KEYS: {'id'},
                self.REPLICATION_METHOD: self.INCREMENTAL,
                self.REPLICATION_KEYS: {'udate'},
                self.OBEYS_START_DATE: True
            },
            'conversion_triggers': {
                self.PRIMARY_KEYS: {'id'},
                self.REPLICATION_METHOD: self.INCREMENTAL,
                self.REPLICATION_KEYS: {'udate'},
                self.OBEYS_START_DATE: True
            },
            'deal_activities': {
                self.PRIMARY_KEYS: {'id'},
                self.REPLICATION_METHOD: self.INCREMENTAL,
                self.REPLICATION_KEYS: {'cdate'},
                self.OBEYS_START_DATE: True
            },
            'deal_group_users': {
                self.PRIMARY_KEYS: {'id'},
                self.REPLICATION_METHOD: self.FULL_TABLE,
                self.OBEYS_START_DATE: False
            },
            'ecommerce_order_activities': {
                self.PRIMARY_KEYS: {'id'},
                self.REPLICATION_METHOD: self.INCREMENTAL,
                self.REPLICATION_KEYS: {'updated_date'},
                self.OBEYS_START_DATE: True
            },
            'email_activities': {
                self.PRIMARY_KEYS: {'id'},
                self.REPLICATION_METHOD: self.INCREMENTAL,
                self.REPLICATION_KEYS: {'tstamp'},
                self.OBEYS_START_DATE: True
            },
            'goals': {
                self.PRIMARY_KEYS: {'id'},
                self.REPLICATION_METHOD: self.FULL_TABLE,
                self.OBEYS_START_DATE: False
            },
            'site_messages': {
                self.PRIMARY_KEYS: {'id'},
                self.REPLICATION_METHOD: self.INCREMENTAL,
                self.REPLICATION_KEYS: {'ldate'},
                self.OBEYS_START_DATE: True
            },
            'sms': {
                self.PRIMARY_KEYS: {'id'},
                self.REPLICATION_METHOD: self.INCREMENTAL,
                self.REPLICATION_KEYS: {'tstamp'},
                self.OBEYS_START_DATE: True
            }
        }


    def expected_check_streams(self):
        """A set of expected stream names"""
        return set(self.expected_metadata().keys())

    def expected_replication_keys(self):
        """return a dictionary with key of table name and value as a set of replication key fields"""

        return {table: properties.get(self.REPLICATION_KEYS, set()) for table, properties
                in self.expected_metadata().items()}

    def expected_primary_keys(self):
        """return a dictionary with key of table name and value as a set of primary key fields"""
        return {table: properties.get(self.PRIMARY_KEYS, set()) for table, properties
                in self.expected_metadata().items()}

    def expected_replication_method(self):
        """return a dictionary with key of table name nd value of replication method"""
        return {table: properties.get(self.REPLICATION_METHOD, set()) for table, properties
                in self.expected_metadata().items()}
        
    def expected_automatic_fields(self):
        """return a dictionary with key of table name and set of value of automatic(primary key and bookmark field) fields"""
        auto_fields = {}
        for k, v in self.expected_metadata().items():
            auto_fields[k] = v.get(self.PRIMARY_KEYS, set()) |  v.get(self.REPLICATION_KEYS, set())
        return auto_fields

    def run_and_verify_check_mode(self, conn_id):
        """
        Run the tap in check mode and verify it succeeds.
        This should be ran prior to field selection and initial sync.
        Return the connection id and found catalogs from menagerie.
        """
        # run in check mode
        check_job_name = runner.run_check_mode(self, conn_id)

        # verify check exit codes
        exit_status = menagerie.get_exit_status(conn_id, check_job_name)
        menagerie.verify_check_exit_status(self, exit_status, check_job_name)

        found_catalogs = menagerie.get_catalogs(conn_id)
        self.assertGreater(len(found_catalogs), 0, msg="unable to locate schemas for connection {}".format(conn_id))

        found_catalog_names = set(map(lambda c: c['stream_name'], found_catalogs))
        self.assertSetEqual(self.expected_check_streams(), found_catalog_names, msg="discovered schemas do not match")
        print("discovered schemas are OK")

        return found_catalogs

    def run_and_verify_sync(self, conn_id):
        """
        Run a sync job and make sure it exited properly.
        Return a dictionary with keys of streams synced
        and values of records synced for each stream
        """

        # Run a sync job using orchestrator
        sync_job_name = runner.run_sync_mode(self, conn_id)

        # verify tap and target exit codes
        exit_status = menagerie.get_exit_status(conn_id, sync_job_name)
        menagerie.verify_sync_exit_status(self, exit_status, sync_job_name)

        sync_record_count = runner.examine_target_output_file(self,
                                                              conn_id,
                                                              self.expected_check_streams(),
                                                              self.expected_primary_keys())

        self.assertGreater(
            sum(sync_record_count.values()), 0,
            msg="failed to replicate any data: {}".format(sync_record_count)
        )
        print("total replicated row count: {}".format(sum(sync_record_count.values())))

        return sync_record_count

    def perform_and_verify_table_and_field_selection(self,
                                                     conn_id,
                                                     test_catalogs,
                                                     select_all_fields=True):
        """
        Perform table and field selection based off of the streams to select
        set and field selection parameters.
        Verify this results in the expected streams selected and all or no
        fields selected for those streams.
        """

        # Select all available fields or select no fields from all testable streams
        self.select_all_streams_and_fields(
            conn_id=conn_id, catalogs=test_catalogs, select_all_fields=select_all_fields
        )

        catalogs = menagerie.get_catalogs(conn_id)

        # Ensure our selection affects the catalog
        expected_selected = [tc.get('stream_name') for tc in test_catalogs]
        for cat in catalogs:
            catalog_entry = menagerie.get_annotated_schema(conn_id, cat['stream_id'])

            # Verify all testable streams are selected
            selected = catalog_entry.get('annotated-schema').get('selected')
            print("Validating selection on {}: {}".format(cat['stream_name'], selected))
            if cat['stream_name'] not in expected_selected:
                self.assertFalse(selected, msg="Stream selected, but not testable.")
                continue # Skip remaining assertions if we aren't selecting this stream
            self.assertTrue(selected, msg="Stream not selected.")

            if select_all_fields:
                # Verify all fields within each selected stream are selected
                for field, field_props in catalog_entry.get('annotated-schema').get('properties').items():
                    field_selected = field_props.get('selected')
                    print("\tValidating selection on {}.{}: {}".format(
                        cat['stream_name'], field, field_selected))
                    self.assertTrue(field_selected, msg="Field not selected.")
            else:
                # Verify only automatic fields are selected
                expected_automatic_fields = self.expected_automatic_fields().get(cat['stream_name'])
                selected_fields = self.get_selected_fields_from_metadata(catalog_entry['metadata'])
                self.assertEqual(expected_automatic_fields, selected_fields)

    @staticmethod
    def get_selected_fields_from_metadata(metadata):
        selected_fields = set()
        for field in metadata:
            is_field_metadata = len(field['breadcrumb']) > 1
            inclusion_automatic_or_selected = (
                field['metadata']['selected'] is True or \
                field['metadata']['inclusion'] == 'automatic'
            )
            if is_field_metadata and inclusion_automatic_or_selected:
                selected_fields.add(field['breadcrumb'][1])
        return selected_fields


    @staticmethod
    def select_all_streams_and_fields(conn_id, catalogs, select_all_fields: bool = True):
        """Select all streams and all fields within streams"""
        for catalog in catalogs:
            schema = menagerie.get_annotated_schema(conn_id, catalog['stream_id'])

            non_selected_properties = []
            if not select_all_fields:
                # get a list of all properties so that none are selected
                non_selected_properties = schema.get('annotated-schema', {}).get(
                    'properties', {}).keys()

            connections.select_catalog_and_fields_via_metadata(
                conn_id, catalog, schema, [], non_selected_properties)

    def calculated_states_by_stream(self, current_state):
        timedelta_by_stream = {stream: [0,0,1]  # {stream_name: [days, hours, minutes], ...}
                               for stream in self.expected_check_streams()}

        stream_to_calculated_state = {stream: "" for stream in current_state['bookmarks'].keys()}
        for stream, state in current_state['bookmarks'].items():
            state_as_datetime = dateutil.parser.parse(state)

            days, hours, minutes = timedelta_by_stream[stream]
            calculated_state_as_datetime = state_as_datetime - timedelta(days=days, hours=hours, minutes=minutes)

            calculated_state_formatted = dt.strftime(calculated_state_as_datetime, self.START_DATE_FORMAT)

            stream_to_calculated_state[stream] = calculated_state_formatted

        return stream_to_calculated_state

    def convert_state_to_utc(self, date_str):
        """
        Convert a saved bookmark value of the form '2020-08-25T13:17:36-07:00' to
        a string formatted utc datetime,
        in order to compare aginast json formatted datetime values
        """
        date_object = dateutil.parser.parse(date_str)
        date_object_utc = date_object.astimezone(tz=pytz.UTC)
        return dt.strftime(date_object_utc, "%Y-%m-%dT%H:%M:%SZ")
    
    def timedelta_formatted(self, dtime, days=0):
        try:
            date_stripped = dt.strptime(dtime, self.START_DATE_FORMAT)
            return_date = date_stripped + timedelta(days=days)

            return dt.strftime(return_date, self.START_DATE_FORMAT)

        except ValueError:
                return Exception("Datetime object is not of the format: {}".format(self.START_DATE_FORMAT))

    def parse_date(self, date_value):
        """
        Pass in string-formatted-datetime, parse the value, and return it as an unformatted datetime object.
        """
        date_formats = {
            "%Y-%m-%dT%H:%M:%S.%fZ",
            "%Y-%m-%dT%H:%M:%SZ",
            "%Y-%m-%dT%H:%M:%S.%f+00:00",
            "%Y-%m-%dT%H:%M:%S+00:00",
            "%Y-%m-%d"
        }
        for date_format in date_formats:
            try:
                date_stripped = dt.strptime(date_value, date_format)
                return date_stripped
            except ValueError:
                continue

        raise NotImplementedError(
            "Tests do not account for dates of this format: {}".format(date_value))



import tap_tester.connections as connections
import tap_tester.runner as runner
import tap_tester.menagerie as menagerie
from base import ActiveCampaignTest


class ActiveCampaignPagination(ActiveCampaignTest):
    """
    Ensure tap can replicate multiple pages of data for streams that use pagination.
    """
    API_LIMIT = 100
    def name(self):
        return "activecampaign_pagination_test"

    def test_run(self):
        """
        • Verify that for each stream you can get multiple pages of data.  
        This requires we ensure more than 1 page of data exists at all times for any given stream.
        • Verify by pks that the data replicated matches the data we expect.
        """

        # Streams to verify all fields tests
        expected_streams = self.expected_check_streams()

        # We are not able to generate enough data to test pagination for below streams,
        # `brandings`, `configs`, `conversions`, `conversion_triggers`, `goals`, `contact_conversions`, `sms`, `user`.
        # The current plan allows only 25 users. So, skipped `user` stream. sms feature is not supported for the current plan.
        # So, removing it all from expected_streams set.
        expected_streams = expected_streams - {'contact_conversions', 'sms', 'users', 'brandings', 'configs', 'conversions', 'conversion_triggers', 'goals'}

        conn_id = connections.ensure_connection(self)

        found_catalogs = self.run_and_verify_check_mode(conn_id)

        # table and field selection
        test_catalogs_all_fields = [catalog for catalog in found_catalogs
                                    if catalog.get('tap_stream_id') in expected_streams]

        self.perform_and_verify_table_and_field_selection(
            conn_id, test_catalogs_all_fields)

        record_count_by_stream = self.run_and_verify_sync(conn_id)

        synced_records = runner.get_records_from_target_output()

        # Verify no unexpected streams were replicated
        synced_stream_names = set(synced_records.keys())
        self.assertSetEqual(expected_streams, synced_stream_names)

        for stream in expected_streams:
            with self.subTest(stream=stream):

                # expected values
                expected_primary_keys = self.expected_primary_keys()[stream]
         
                # verify that we can paginate with all fields selected
                record_count_sync = record_count_by_stream.get(stream, 0)
                self.assertGreater(record_count_sync, self.API_LIMIT,
                                    msg="The number of records is not over the stream max limit")

                primary_keys_list = [tuple([message.get('data').get(expected_pk) for expected_pk in expected_primary_keys])
                                        for message in synced_records.get(stream).get('messages')
                                        if message.get('action') == 'upsert']

                primary_keys_list_1 = primary_keys_list[:self.API_LIMIT]
                primary_keys_list_2 = primary_keys_list[self.API_LIMIT:2*self.API_LIMIT]

                primary_keys_page_1 = set(primary_keys_list_1)
                primary_keys_page_2 = set(primary_keys_list_2)

                # Verify by primary keys that data is unique for page
                self.assertTrue(
                    primary_keys_page_1.isdisjoint(primary_keys_page_2))


import tap_tester.connections as connections
import tap_tester.runner as runner
from base import ActiveCampaignTest

class ActiveCampaignAutomaticFields(ActiveCampaignTest):
    """
    Ensure running the tap with all streams selected and all fields deselected results in the replication of just the 
    primary keys and replication keys (automatic fields).
    """
    
    def name(self):
        return "activecampaign_automatic_fields"

    def test_run(self):
        """
        Verify we can deselect all fields except when inclusion=automatic, which is handled by base.py methods
        Verify that only the automatic fields are sent to the target.
        Verify that all replicated records have unique primary key values.
        """
        
        streams_to_test = self.expected_check_streams()

        # We are not able to generate data for `contact_conversions` stream.
        # For `sms` it requires Enterprise plan of account. So, removing it from streams_to_test set.
        streams_to_test = streams_to_test - {'contact_conversions', 'sms'}
        
        conn_id = connections.ensure_connection(self)

        found_catalogs = self.run_and_verify_check_mode(conn_id)

        # table and field selection
        test_catalogs_automatic_fields = [catalog for catalog in found_catalogs
                                          if catalog.get('tap_stream_id') in streams_to_test]

        # Select all streams and no fields within streams
        self.perform_and_verify_table_and_field_selection(
            conn_id, test_catalogs_automatic_fields, select_all_fields=False)

        record_count_by_stream = self.run_and_verify_sync(conn_id)
        synced_records = runner.get_records_from_target_output()
        
        for stream in streams_to_test:
            with self.subTest(stream=stream):

                # expected values
                expected_keys = self.expected_automatic_fields().get(stream)
                expected_primary_keys = self.expected_primary_keys()[stream]
                
                # collect actual values
                data = synced_records.get(stream, {})
                record_messages_keys = [set(row['data'].keys())
                                        for row in data.get('messages', [])]
                primary_keys_list = [tuple(message.get('data', {}).get(expected_pk) for expected_pk in expected_primary_keys)
                                       for message in data.get('messages', [])
                                       if message.get('action') == 'upsert']
                unique_primary_keys_list = set(primary_keys_list)
                
                # Verify that you get some records for each stream
                self.assertGreater(
                    record_count_by_stream.get(stream, -1), 0,
                    msg="The number of records is not over the stream min limit")

                # Verify that only the automatic fields are sent to the target
                for actual_keys in record_messages_keys:
                    self.assertSetEqual(expected_keys, actual_keys)
                    
                #Verify that all replicated records have unique primary key values.
                self.assertEqual(len(primary_keys_list), 
                                    len(unique_primary_keys_list), 
                                    msg="Replicated record does not have unique primary key values.")


import re
import singer
import humps

LOGGER = singer.get_logger()


def fix_records(this_json):
    new_json = []
    for rec in this_json:
        if 'links' in rec:
            if isinstance(rec['links'], dict):
                rec.pop('links', None)
        for key, val in rec.items():
            if 'date' in key or 'stamp' in key or key in ('socialdata_lastcheck', 'deleted_at'):
                if val == '0000-00-00 00:00:00':
                    rec[key] = None

        new_json.append(rec)


def transform_json(this_json, stream_name, data_key):
    if data_key in this_json:
        converted_json = humps.decamelize(this_json[data_key])
    else:
        converted_json = humps.decamelize(this_json)
    
    fix_records(converted_json)

    return converted_json



import tap_tester.connections as connections
import tap_tester.runner as runner
from base import ActiveCampaignTest
from tap_tester import menagerie
from datetime import datetime
import uuid
import os
import time

class ActiveCampaignBookMark(ActiveCampaignTest):
    """Test tap sets a bookmark and respects it for the next sync of a stream"""
    
    def name(self):
        return "activecampaign_bookmark_test"

    def test_run(self):
        """
        Verify that for each stream you can do a sync which records bookmarks.
        That the bookmark is the maximum value sent to the target for the replication key.
        That a second sync respects the bookmark
            All data of the second sync is >= the bookmark from the first sync
            The number of records in the 2nd sync is less then the first (This assumes that
                new data added to the stream is done at a rate slow enough that you haven't
                doubled the amount of data from the start date to the first sync between
                the first sync and second sync run in this test)
        Verify that for full table stream, all data replicated in sync 1 is replicated again in sync 2.
        PREREQUISITE
        For EACH stream that is incrementally replicated there are multiple rows of data with
            different values for the replication key
        """
        
        
        expected_streams = self.expected_check_streams()
        # We are not able to generate data for `contact_conversions` stream.
        # For `sms` stream it requires Enterprise plan of account. So, removing it from expected_streams set.
        expected_streams = expected_streams - {'contact_conversions', 'sms'}

        expected_replication_keys = self.expected_replication_keys()
        expected_replication_methods = self.expected_replication_method()

        ##########################################################################
        # First Sync
        ##########################################################################
        conn_id = connections.ensure_connection(self)

        # Run in check mode
        found_catalogs = self.run_and_verify_check_mode(conn_id)

        # table and field selection
        catalog_entries = [catalog for catalog in found_catalogs
                           if catalog.get('tap_stream_id') in expected_streams]

        self.perform_and_verify_table_and_field_selection(
            conn_id, catalog_entries)

        # Run a first sync job using orchestrator
        first_sync_record_count = self.run_and_verify_sync(conn_id)
        first_sync_records = runner.get_records_from_target_output()
        first_sync_bookmarks = menagerie.get_state(conn_id)

        ##########################################################################
        # Update State Between Syncs
        ##########################################################################

        new_states = {'bookmarks': dict()}
        simulated_states = self.calculated_states_by_stream(
            first_sync_bookmarks)
        for stream, new_state in simulated_states.items():
            new_states['bookmarks'][stream] = new_state
        menagerie.set_state(conn_id, new_states)

        ##########################################################################
        # Second Sync
        ##########################################################################

        second_sync_record_count = self.run_and_verify_sync(conn_id)
        second_sync_records = runner.get_records_from_target_output()
        second_sync_bookmarks = menagerie.get_state(conn_id)

        ##########################################################################
        # Test By Stream
        ##########################################################################


        for stream in expected_streams:
            with self.subTest(stream=stream):

                # expected values
                expected_replication_method = expected_replication_methods[stream]

                # collect information for assertions from syncs 1 & 2 base on expected values
                first_sync_count = first_sync_record_count.get(stream, 0)
                second_sync_count = second_sync_record_count.get(stream, 0)
                first_sync_messages = [record.get('data') for record in
                                       first_sync_records.get(
                                           stream, {}).get('messages', [])
                                       if record.get('action') == 'upsert']
                second_sync_messages = [record.get('data') for record in
                                        second_sync_records.get(
                                            stream, {}).get('messages', [])
                                        if record.get('action') == 'upsert']
                first_bookmark_value = first_sync_bookmarks.get('bookmarks', {stream: None}).get(stream)
                second_bookmark_value = second_sync_bookmarks.get('bookmarks', {stream: None}).get(stream)
                

                if expected_replication_method == self.INCREMENTAL:

                    # collect information specific to incremental streams from syncs 1 & 2
                    replication_key = next(
                        iter(expected_replication_keys[stream]))
                    first_bookmark_value_utc = self.convert_state_to_utc(
                        first_bookmark_value)
                    second_bookmark_value_utc = self.convert_state_to_utc(
                        second_bookmark_value)

                    
                    simulated_bookmark_value = self.convert_state_to_utc(new_states['bookmarks'][stream])
                    
                    # Verify the first sync sets a bookmark of the expected form
                    self.assertIsNotNone(first_bookmark_value)

                    # Verify the second sync sets a bookmark of the expected form
                    self.assertIsNotNone(second_bookmark_value)

                    # Verify the second sync bookmark is Equal to the first sync bookmark
                    # assumes no changes to data during test
                    self.assertEqual(second_bookmark_value, first_bookmark_value)

                    for record in first_sync_messages:

                        # Verify the first sync bookmark value is the max replication key value for a given stream
                        replication_key_value = record.get(replication_key)
   
                        self.assertLessEqual(
                            replication_key_value, first_bookmark_value_utc,
                            msg="First sync bookmark was set incorrectly, a record with a greater replication-key value was synced."
                        )
                    
                    for record in second_sync_messages:
                        # Verify the second sync replication key value is Greater or Equal to the first sync bookmark
                        replication_key_value = record.get(replication_key)

                        self.assertGreaterEqual(replication_key_value, simulated_bookmark_value,
                                                msg="Second sync records do not repect the previous bookmark.")

                        # Verify the second sync bookmark value is the max replication key value for a given stream
                        self.assertLessEqual(
                            replication_key_value, second_bookmark_value_utc,
                            msg="Second sync bookmark was set incorrectly, a record with a greater replication-key value was synced."
                        )

                elif expected_replication_method == self.FULL_TABLE:

                    # Verify the syncs do not set a bookmark for full table streams
                    self.assertIsNone(first_bookmark_value)
                    self.assertIsNone(first_bookmark_value)

                    # Verify the number of records in the second sync is the same as the first
                    self.assertEqual(second_sync_count, first_sync_count)

                else:

                    raise NotImplementedError(
                        "INVALID EXPECTATIONS\t\tSTREAM: {} REPLICATION_METHOD: {}".format(
                            stream, expected_replication_method)
                    )

                # Verify at least 1 record was replicated in the second sync
                self.assertGreater(
                    second_sync_count, 0, msg="We are not fully testing bookmarking for {}".format(stream))